# Rate Limits

This document describes the rate limits for the CoinsSend API and how to handle rate limiting in your applications.

## Rate Limit Overview

The CoinsSend API implements rate limiting to ensure fair usage and protect the system from abuse. Rate limits are applied at two levels:

1. **Per Merchant**: 400 requests per minute per merchant ID
2. **Per IP Address**: 20 requests per second per IP address

When a rate limit is exceeded, the API will return a `429 Too Many Requests` HTTP status code.

## Rate Limit Headers

When making API requests, you'll receive headers that provide information about your current rate limit status:

| Header | Description |
|--------|-------------|
| `X-RateLimit-Limit` | The maximum number of requests allowed in the current time window |
| `X-RateLimit-Remaining` | The number of requests remaining in the current time window |
| `X-RateLimit-Reset` | The time (in Unix timestamp seconds) when the rate limit window resets |

Example response headers:

```
X-RateLimit-Limit: 400
X-RateLimit-Remaining: 395
X-RateLimit-Reset: 1715123456
```

## Rate Limit Response

When you exceed the rate limit, you'll receive a response with HTTP status code 429 and a JSON body like:

```json
{
  "status": "error",
  "error": {
    "code": 429,
    "message": "Too many requests. Please try again in 35 seconds."
  }
}
```

## Best Practices for Handling Rate Limits

### 1. Implement Exponential Backoff

When you receive a 429 response, implement exponential backoff with jitter to retry the request:

<div datatype="code-tabs">

```php
// PHP Example
function makeRequestWithBackoff($client, $method, $url, $options, $maxRetries = 5) {
    $attempts = 0;
    
    while ($attempts < $maxRetries) {
        try {
            $response = $client->request($method, $url, $options);
            return json_decode($response->getBody(), true);
            
        } catch (ClientException $e) {
            if ($e->getResponse()->getStatusCode() === 429) {
                $attempts++;
                
                // Get retry-after header or use exponential backoff
                $retryAfter = $e->getResponse()->getHeader('Retry-After')[0] ?? null;
                
                if ($retryAfter) {
                    // Server specified retry delay
                    $delay = (int)$retryAfter;
                } else {
                    // Exponential backoff with jitter
                    $delay = (2 ** $attempts) + rand(0, 1000) / 1000;
                }
                
                error_log("Rate limited. Retrying in {$delay} seconds (attempt {$attempts}/{$maxRetries})");
                sleep($delay);
                continue;
            }
            
            // Re-throw non-rate-limit errors
            throw $e;
        }
    }
    
    // Max retries reached
    throw new Exception("Rate limit exceeded. Max retries reached.");
}
```

```javascript
// JavaScript Example
async function makeRequestWithBackoff(url, options, maxRetries = 5) {
  let attempts = 0;
  
  while (attempts < maxRetries) {
    try {
      const response = await fetch(url, options);
      
      // Check for rate limit error
      if (response.status === 429) {
        attempts++;
        
        // Get retry-after header or use exponential backoff
        const retryAfter = response.headers.get('Retry-After');
        
        // Calculate delay (in milliseconds)
        let delay;
        if (retryAfter) {
          // Server specified retry delay
          delay = parseInt(retryAfter, 10) * 1000;
        } else {
          // Exponential backoff with jitter
          delay = (Math.pow(2, attempts) * 1000) + (Math.random() * 1000);
        }
        
        console.warn(`Rate limited. Retrying in ${delay/1000} seconds (attempt ${attempts}/${maxRetries})`);
        await new Promise(resolve => setTimeout(resolve, delay));
        continue;
      }
      
      // Parse and return response for non-rate-limit responses
      return await response.json();
      
    } catch (error) {
      // Handle network errors
      throw error;
    }
  }
  
  // Max retries reached
  throw new Error("Rate limit exceeded. Max retries reached.");
}
```

```python
# Python Example
import time
import random
import requests

def make_request_with_backoff(url, method='GET', data=None, headers=None, max_retries=5):
    attempts = 0
    
    while attempts < max_retries:
        try:
            response = requests.request(method, url, json=data, headers=headers)
            
            # Check for rate limit error
            if response.status_code == 429:
                attempts += 1
                
                # Get retry-after header or use exponential backoff
                retry_after = response.headers.get('Retry-After')
                
                if retry_after:
                    # Server specified retry delay
                    delay = int(retry_after)
                else:
                    # Exponential backoff with jitter
                    delay = (2 ** attempts) + random.random()
                
                print(f"Rate limited. Retrying in {delay} seconds (attempt {attempts}/{max_retries})")
                time.sleep(delay)
                continue
            
            # Return response for non-rate-limit responses
            return response
            
        except requests.exceptions.RequestException as e:
            # Handle request exceptions
            raise e
    
    # Max retries reached
    raise Exception("Rate limit exceeded. Max retries reached.")
```

</div>

### 2. Monitor Rate Limit Headers

Track the rate limit headers in your responses to understand your usage:

<div datatype="code-tabs">

```php
// PHP Example
function checkRateLimitHeaders($response) {
    $limit = $response->getHeader('X-RateLimit-Limit')[0] ?? null;
    $remaining = $response->getHeader('X-RateLimit-Remaining')[0] ?? null;
    $reset = $response->getHeader('X-RateLimit-Reset')[0] ?? null;
    
    if ($limit && $remaining) {
        $usagePercentage = 100 - (($remaining / $limit) * 100);
        
        // Log when usage is high
        if ($usagePercentage > 80) {
            error_log("WARNING: API rate limit usage at {$usagePercentage}%. {$remaining}/{$limit} requests remaining.");
        }
        
        // Return usage info
        return [
            'limit' => (int)$limit,
            'remaining' => (int)$remaining,
            'reset' => $reset ? new DateTime("@$reset") : null,
            'usage_percentage' => $usagePercentage
        ];
    }
    
    return null;
}
```

```javascript
// JavaScript Example
function checkRateLimitHeaders(response) {
  const limit = parseInt(response.headers.get('X-RateLimit-Limit'), 10);
  const remaining = parseInt(response.headers.get('X-RateLimit-Remaining'), 10);
  const reset = parseInt(response.headers.get('X-RateLimit-Reset'), 10);
  
  if (limit && remaining) {
    const usagePercentage = 100 - ((remaining / limit) * 100);
    
    // Log when usage is high
    if (usagePercentage > 80) {
      console.warn(`WARNING: API rate limit usage at ${usagePercentage.toFixed(2)}%. ${remaining}/${limit} requests remaining.`);
    }
    
    // Return usage info
    return {
      limit,
      remaining,
      reset: reset ? new Date(reset * 1000) : null,
      usagePercentage
    };
  }
  
  return null;
}
```

```python
# Python Example
def check_rate_limit_headers(response):
    limit = response.headers.get('X-RateLimit-Limit')
    remaining = response.headers.get('X-RateLimit-Remaining')
    reset = response.headers.get('X-RateLimit-Reset')
    
    if limit and remaining:
        limit = int(limit)
        remaining = int(remaining)
        usage_percentage = 100 - ((remaining / limit) * 100)
        
        # Log when usage is high
        if usage_percentage > 80:
            print(f"WARNING: API rate limit usage at {usage_percentage:.2f}%. {remaining}/{limit} requests remaining.")
        
        # Return usage info
        return {
            'limit': limit,
            'remaining': remaining,
            'reset': datetime.fromtimestamp(int(reset)) if reset else None,
            'usage_percentage': usage_percentage
        }
    
    return None
```

</div>

### 3. Implement Request Queuing

For high-volume applications, implement a request queue with controlled concurrency:

<div datatype="code-tabs">

```php
// PHP Example with Request Queuing
class ApiRequestQueue {
    private $queue = [];
    private $processing = false;
    private $requestsPerSecond;
    private $lastRequestTime;
    
    public function __construct($requestsPerSecond = 3) {
        $this->requestsPerSecond = $requestsPerSecond;
        $this->lastRequestTime = 0;
    }
    
    public function addRequest($callback) {
        $this->queue[] = $callback;
        
        if (!$this->processing) {
            $this->processQueue();
        }
    }
    
    private function processQueue() {
        $this->processing = true;
        
        while (!empty($this->queue)) {
            // Respect rate limits
            $this->throttle();
            
            $callback = array_shift($this->queue);
            $this->lastRequestTime = microtime(true);
            
            try {
                $callback();
            } catch (Exception $e) {
                error_log("API Request error: " . $e->getMessage());
                // Handle rate limit here
                if ($e->getCode() === 429) {
                    // Requeue and delay processing
                    array_unshift($this->queue, $callback);
                    sleep(1);
                }
            }
        }
        
        $this->processing = false;
    }
    
    private function throttle() {
        $timeSinceLastRequest = microtime(true) - $this->lastRequestTime;
        $minRequestInterval = 1 / $this->requestsPerSecond;
        
        if ($timeSinceLastRequest < $minRequestInterval) {
            usleep(($minRequestInterval - $timeSinceLastRequest) * 1000000);
        }
    }
}
```

```javascript
// JavaScript Example with Request Queuing
class ApiRequestQueue {
  constructor(requestsPerSecond = 3) {
    this.queue = [];
    this.processing = false;
    this.requestsPerSecond = requestsPerSecond;
    this.lastRequestTime = 0;
  }
  
  addRequest(requestFunc) {
    return new Promise((resolve, reject) => {
      this.queue.push({
        requestFunc,
        resolve,
        reject
      });
      
      if (!this.processing) {
        this.processQueue();
      }
    });
  }
  
  async processQueue() {
    this.processing = true;
    
    while (this.queue.length > 0) {
      // Respect rate limits
      await this.throttle();
      
      const { requestFunc, resolve, reject } = this.queue.shift();
      this.lastRequestTime = Date.now();
      
      try {
        const result = await requestFunc();
        resolve(result);
      } catch (error) {
        // Handle rate limit
        if (error.status === 429) {
          // Get retry time from header or default to 1 second
          const retryAfter = parseInt(error.headers?.get('Retry-After'), 10) || 1;
          console.warn(`Rate limited. Waiting ${retryAfter}s before retrying.`);
          
          // Put the request back at the front of the queue
          this.queue.unshift({ requestFunc, resolve, reject });
          await new Promise(r => setTimeout(r, retryAfter * 1000));
        } else {
          reject(error);
        }
      }
    }
    
    this.processing = false;
  }
  
  async throttle() {
    const now = Date.now();
    const timeSinceLastRequest = now - this.lastRequestTime;
    const minRequestInterval = 1000 / this.requestsPerSecond;
    
    if (timeSinceLastRequest < minRequestInterval) {
      await new Promise(resolve => 
        setTimeout(resolve, minRequestInterval - timeSinceLastRequest)
      );
    }
  }
}
```

```python
# Python Example with Request Queuing
import time
import threading
import queue
import logging
from typing import Callable, Any

class ApiRequestQueue:
    def __init__(self, requests_per_second=3):
        self.queue = queue.Queue()
        self.requests_per_second = requests_per_second
        self.last_request_time = 0
        self.processing = False
        self.thread = None
    
    def add_request(self, request_func):
        future = threading.Event()
        result = {'data': None, 'error': None}
        
        self.queue.put((request_func, future, result))
        
        if not self.processing:
            self.process_queue()
        
        # Wait for request to complete
        future.wait()
        
        if result['error']:
            raise result['error']
        return result['data']
    
    def process_queue(self):
        if self.processing:
            return
            
        self.processing = True
        self.thread = threading.Thread(target=self._process_worker)
        self.thread.daemon = True
        self.thread.start()
    
    def _process_worker(self):
        while not self.queue.empty():
            # Respect rate limits
            self._throttle()
            
            request_func, future, result = self.queue.get()
            self.last_request_time = time.time()
            
            try:
                result['data'] = request_func()
                future.set()
            except Exception as e:
                # Handle rate limit
                if getattr(e, 'status_code', 0) == 429:
                    # Get retry time from header or default to 1 second
                    retry_after = getattr(e, 'headers', {}).get('Retry-After', 1)
                    retry_after = int(retry_after) if retry_after else 1
                    logging.warning(f"Rate limited. Waiting {retry_after}s before retrying.")
                    
                    # Put the request back in the queue
                    self.queue.put((request_func, future, result))
                    time.sleep(retry_after)
                else:
                    result['error'] = e
                    future.set()
            
            self.queue.task_done()
        
        self.processing = False
    
    def _throttle(self):
        now = time.time()
        time_since_last_request = now - self.last_request_time
        min_request_interval = 1.0 / self.requests_per_second
        
        if time_since_last_request < min_request_interval:
            time.sleep(min_request_interval - time_since_last_request)
```

</div>

### 4. Optimize API Usage

Reduce your API usage by implementing these strategies:

1. **Caching**: Cache responses for endpoints that don't frequently change
2. **Batch Requests**: Combine multiple operations into single requests when possible
3. **Conditional Requests**: Use If-Modified-Since headers to avoid retrieving unchanged data
4. **Pagination**: Use smaller page sizes for listing endpoints

## Rate Limit Strategies by Endpoint

Different endpoints have different rate limit considerations:

| Endpoint | Rate Limit Consideration |
|----------|--------------------------|
| `POST /v1/invoices` | High-priority endpoint; implement request queuing with retries |
| `POST /v1/withdrawals` | Critical endpoint; use exponential backoff with extended retry attempts |

## Monitoring Rate Limit Usage

Set up monitoring to detect when you're approaching rate limits:

1. **Log Rate Limit Headers**: Track the X-RateLimit-* headers with each response
2. **Set Alerts**: Trigger alerts when rate limit usage consistently exceeds 80%
3. **Track 429 Responses**: Monitor the frequency of rate limit errors

## Rate Limit Increase Requests

If your application legitimately requires higher rate limits, you can request an increase by contacting Sphere support. Be prepared to provide:

1. Your merchant ID
2. The specific endpoints needing increased limits
3. Justification for the increased limit requirement
4. Expected request volume

## FAQ

### Q: How can I avoid hitting rate limits?

- Implement request queuing with controlled concurrency
- Cache API responses where appropriate
- Use bulk operations instead of multiple individual requests
- Add jitter to retry attempts to prevent synchronization

### Q: What should I do when I receive a 429 response?

- Respect the Retry-After header if provided
- Use exponential backoff with jitter for retries
- Implement request queuing to control request rates
- Consider reducing your application's API usage by implementing caching

### Q: Do different endpoints have different rate limits?

No, the rate limits are applied globally across all endpoints. However, some endpoints (like webhooks) are exempt from rate limits.